"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});


/**
 * Regex to match urls ending with "?args=value" and to make it
 * match precise routes.
 *
 * @type {string}
 */

/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

var URLArgsRegex = "((\\?)([^=]+)(=(.+))?)?$";

/**
 * Return true if the url is matching the route
 *
 * @param route
 * @param url
 * @returns {boolean}
 */
var isMatching = exports.isMatching = function isMatching(route, url) {
  var urlRegex = route.replace(/{(\w)+}/g, "([^/]+)") + URLArgsRegex;
  return new RegExp(urlRegex).test(url);
};

/**
 * return an object containing the path and the args of an URL
 * e.g : site.com/index?foo=bar would make { uri: site.com/index, args: foo=bar }
 *
 * @param url
 * @returns {{uri: string, args: string}}
 */
var splitPathAndArgs = exports.splitPathAndArgs = function splitPathAndArgs(url) {
  var splitUrl = url.split('?');

  var output = {
    uri: splitUrl[0],
    args: undefined
  };

  if (splitUrl[1]) {
    output.args = splitUrl[1];
  }

  return output;
};

/**
 * Extract all the arguments from the parameters, following the ? in the URL
 * @param url
 * @returns {{}}
 */
var extractURLArgs = exports.extractURLArgs = function extractURLArgs(url) {
  var urlObject = splitPathAndArgs(url),
      args = {};

  if (urlObject.args) {
    var pairs = urlObject.args.split('&');
    if (pairs.length === 0) {
      return args;
    }

    [].slice.call(pairs).forEach(function (pair) {
      var keyValue = pair.split("=");
      args[keyValue[0]] = keyValue[1];
    });

    return args;
  }

  return {};
};

/**
 * Apply the template to the url to return captured values
 *
 * From template such as http://url.com/post/{id}
 * with url such as http://url.com/post/3
 * This function will match ID with the value 3.
 *
 * @param template
 * @param url
 * @returns {{}}
 */
var applyTemplateToUrl = exports.applyTemplateToUrl = function applyTemplateToUrl(template, url) {
  var keys = [];

  // Store every key, and return a regex instead.
  var urlRegex = template.replace(/{(\w)+}/g, function (arg) {
    keys.push(arg.substr(1, arg.length - 2));
    return "([^?]+)";
  });

  var regex = new RegExp(urlRegex);
  var routeArgs = regex.exec(url);

  var obj = {};
  [].slice.call(routeArgs, 1).forEach(function (value, index) {
    obj[keys[index]] = value;
  });

  return obj;
};

/**
 * Mutate the request to enhance it :
 * - pattern matching : match site.com/foo/{id} and site.com/foo/1 to return { id: 1 } in the route bag
 * - extract URL arguments (foo=bar&qux) into the queryBag
 * - fill in the request bag
 * @param request
 * @param url
 */
function enhanceRequest(request, url) {
  var queryBag = Object.assign({}, extractURLArgs(request.url), request.query),
      requestBag = {},
      routeBag = applyTemplateToUrl(url, request.url);

  // In case of anything but a GET request, we fill the request variable with the body's data
  if (request.method !== "GET") {
    requestBag = request.request;
  }

  request.query = queryBag;
  request.request = requestBag;
  request.route = routeBag;
}

/**
 * @module Faussaire
 * Mock API calls responses.
 *
 * Faussaire allows to preconfigure a response. The user can then make a fake "request". An internal router will
 * associate the URL and Method given to the preconfigured response.
 * The preconfigured response can be configured. It can be "consumable" which means after the router has matched an URL
 * with the response, it is removed from it.
 *
 * The Waiters is a queue. We look in reverse mode : that is, the last waiter registered is the first to be checked.
 *
 * This is how the algorithm works :
 * - We call the "emit" function with an URL, a method, and bunch of configuration like headers, timeout, etc
 * - we look for a preconfigured response whose waiting for this URL and this method
 * - Once we have a route, we check his "action"
 *   - if it's an "execute", we execute the function called "trigger", and we return it's output
 *   - if it's a "timeout", we return a timeout
 * - If we don't have a return, we output a standard 404-like error
 */
function Faussaire() {
  this.waiters = [];
}

/**
 * Add a waiter to the queue
 * @param config
 * @returns {Faussaire}
 */
Faussaire.prototype.observe = function (config) {
  var nextWaiter = Object.assign({}, {
    consumable: true
  }, config);

  // We add the element at the top of the array because faussaire is LIFO
  this.waiters.unshift(nextWaiter);
  return this;
};

/**
 * Forward a request to the last registered waiter and return a result
 * @param request
 * @returns {Result}
 */
Faussaire.prototype.emit = function (request) {
  var waiter = null,
      response = null,
      result = null;

  waiter = this.waiters.find(function (waiter) {
    return isMatching(waiter.url, request.url) && waiter.method === request.method;
  });

  if (waiter) {
    if (typeof waiter.action === "function") {
      enhanceRequest(request, waiter.url);

      response = waiter.action(request);
      result = { status: "done", response: response };
    } else if (waiter.action === "timeout") {
      result = Faussaire.createTimeoutResult();
    } else {
      result = Faussaire.createNetworkErrorResult();
    }

    if (waiter.consumable === true) {
      this.waiters = this.waiters.filter(function (_waiter) {
        return _waiter !== waiter;
      });
    }
  } else {
    result = Faussaire.createNotFoundResult();
  }

  return result;
};

var defaultResults = {
  timeout: {
    response: null,
    status: "timeout"
  },
  networkError: {
    response: null,
    status: "network-error"
  },
  notFound: {
    response: {
      statusCode: 404,
      data: null,
      headers: null
    },
    status: "done"
  }
};

/**
 * Create a generic timeout result
 * @returns {{response: null, status: string}}
 */
Faussaire.createTimeoutResult = function () {
  return defaultResults.timeout;
};

/**
 * Create a generic timeout result
 * @returns {{response: null, status: string}}
 */
Faussaire.createNetworkErrorResult = function () {
  return defaultResults.networkError;
};

/**
 * Creates a generic not found result
 * @returns {{response: {statusCode: number, data: null, headers: null}, status: string}}
 */
Faussaire.createNotFoundResult = function () {
  return defaultResults.notFound;
};

exports.default = {
  /**
   * @returns {Faussaire}
   */
  create: function create() {
    return new Faussaire();
  }
};